@@ -58,6 +58,11 @@ Street, Fifth Floor, Boston, MA 02110-1335 USA
 #include "backup_mysql.h"
 #include <btr0btr.h>
 
+#ifdef _WIN32
+#include <aclapi.h>
+#endif
+
+
 #define ROCKSDB_BACKUP_DIR "#rocksdb"
 
 /* list of files to sync for --rsync mode */
@@ -124,7 +129,6 @@ struct datadir_thread_ctxt_t {
 	uint			n_thread;
 	uint			*count;
 	pthread_mutex_t*	count_mutex;
-	os_thread_id_t		id;
 	bool			ret;
 };
 
@@ -269,7 +273,6 @@ datadir_iter_next_database(datadir_iter_t *it)
 		}
 		snprintf(it->dbpath, it->dbpath_len, "%s/%s",
 			 it->datadir_path, it->dbinfo.name);
-		os_normalize_path(it->dbpath);
 
 		if (it->dbinfo.type == OS_FILE_TYPE_FILE) {
 			it->is_file = true;
@@ -465,14 +468,13 @@ struct datafile_cur_t {
 	char		abs_path[FN_REFLEN];
 	MY_STAT		statinfo;
 	uint		thread_n;
-	byte*		orig_buf;
 	byte*		buf;
 	size_t		buf_size;
 	size_t		buf_read;
 	size_t		buf_offset;
 
 	explicit datafile_cur_t(const char* filename = NULL) :
-		file(), thread_n(0), orig_buf(NULL), buf(NULL), buf_size(0),
+		file(), thread_n(0), buf(NULL), buf_size(0),
 		buf_read(0), buf_offset(0)
 	{
 		memset(rel_path, 0, sizeof rel_path);
@@ -578,6 +580,7 @@ datafile_read(datafile_cur_t *cursor)
 Check to see if a file exists.
 Takes name of the file to check.
 @return true if file exists. */
+static
 bool
 file_exists(const char *filename)
 {
@@ -941,7 +944,7 @@ backup_file_printf(const char *filename, const char *fmt, ...)
 
 static
 bool
-run_data_threads(datadir_iter_t *it, os_thread_func_t func, uint n)
+run_data_threads(datadir_iter_t *it, void (*func)(datadir_thread_ctxt_t *ctxt), uint n)
 {
 	datadir_thread_ctxt_t	*data_threads;
 	uint			i, count;
@@ -959,12 +962,12 @@ run_data_threads(datadir_iter_t *it, os_thread_func_t func, uint n)
 		data_threads[i].n_thread = i + 1;
 		data_threads[i].count = &count;
 		data_threads[i].count_mutex = &count_mutex;
-		os_thread_create(func, data_threads + i, &data_threads[i].id);
+		std::thread(func, data_threads + i).detach();
 	}
 
 	/* Wait for threads to exit */
 	while (1) {
-		os_thread_sleep(100000);
+		std::this_thread::sleep_for(std::chrono::milliseconds(100));
 		pthread_mutex_lock(&count_mutex);
 		if (count == 0) {
 			pthread_mutex_unlock(&count_mutex);
@@ -988,64 +991,6 @@ run_data_threads(datadir_iter_t *it, os_thread_func_t func, uint n)
 	return(ret);
 }
 
-#ifdef _WIN32
-#include <windows.h>
-#include <accctrl.h>
-#include <aclapi.h>
-/*
-  On Windows, fix permission of the file after "copyback"
-  We assume that after copyback, mysqld will run as service as NetworkService
-  user, thus well give full permission on given file to that user.
-*/
-
-static int fix_win_file_permissions(const char *file)
-{
-	struct {
-		TOKEN_USER tokenUser;
-		BYTE buffer[SECURITY_MAX_SID_SIZE];
-	} tokenInfoBuffer;
-	HANDLE hFile = CreateFile(file, READ_CONTROL | WRITE_DAC, 0, NULL, OPEN_EXISTING,
-		FILE_FLAG_BACKUP_SEMANTICS, NULL);
-	if (hFile == INVALID_HANDLE_VALUE)
-		return -1;
-	ACL* pOldDACL;
-	SECURITY_DESCRIPTOR* pSD = NULL;
-	EXPLICIT_ACCESS ea = { 0 };
-	PSID pSid = NULL;
-
-	GetSecurityInfo(hFile, SE_FILE_OBJECT, DACL_SECURITY_INFORMATION, NULL, NULL,
-		&pOldDACL, NULL, (void**)&pSD);
-	DWORD size = SECURITY_MAX_SID_SIZE;
-	pSid = (PSID)tokenInfoBuffer.buffer;
-	if (!CreateWellKnownSid(WinNetworkServiceSid, NULL, pSid,
-		&size))
-	{
-		return 1;
-	}
-	ea.Trustee.TrusteeForm = TRUSTEE_IS_SID;
-	ea.Trustee.ptstrName = (LPTSTR)pSid;
-
-	ea.grfAccessMode = GRANT_ACCESS;
-	ea.grfAccessPermissions = GENERIC_ALL;
-	ea.grfInheritance = CONTAINER_INHERIT_ACE | OBJECT_INHERIT_ACE;
-	ea.Trustee.TrusteeType = TRUSTEE_IS_UNKNOWN;
-	ACL* pNewDACL = 0;
-	SetEntriesInAcl(1, &ea, pOldDACL, &pNewDACL);
-	if (pNewDACL)
-	{
-		SetSecurityInfo(hFile, SE_FILE_OBJECT, DACL_SECURITY_INFORMATION, NULL, NULL,
-			pNewDACL, NULL);
-	}
-	if (pSD != NULL)
-		LocalFree((HLOCAL)pSD);
-	if (pNewDACL != NULL)
-		LocalFree((HLOCAL)pNewDACL);
-	CloseHandle(hFile);
-	return 0;
-}
-
-#endif
-
 
 /************************************************************************
 Copy file for backup/restore.
@@ -1096,10 +1041,6 @@ copy_file(ds_ctxt_t *datasink,
 	/* close */
 	msg(thread_n,"        ...done");
 	datafile_close(&cursor);
-#ifdef _WIN32
-	if (xtrabackup_copy_back || xtrabackup_move_back)
-		ut_a(!fix_win_file_permissions(dstfile->path));
-#endif
 	if (ds_close(dstfile)) {
 		goto error_close;
 	}
@@ -1172,10 +1113,6 @@ move_file(ds_ctxt_t *datasink,
 			errbuf);
 		return(false);
 	}
-#ifdef _WIN32
-	if (xtrabackup_copy_back || xtrabackup_move_back)
-		ut_a(!fix_win_file_permissions(dst_file_path_abs));
-#endif
 	msg(thread_n,"        ...done");
 
 	return(true);
@@ -1198,13 +1135,12 @@ read_link_file(const char *ibd_filepath, const char *link_filepath)
 		os_file_read_string(file, filepath, OS_FILE_MAX_PATH);
 		fclose(file);
 
-		if (strlen(filepath)) {
+		if (size_t len = strlen(filepath)) {
 			/* Trim whitespace from end of filepath */
-			ulint lastch = strlen(filepath) - 1;
+			ulint lastch = len - 1;
 			while (lastch > 4 && filepath[lastch] <= 0x20) {
 				filepath[lastch--] = 0x00;
 			}
-			os_normalize_path(filepath);
 		}
 
 		tablespace_locations[ibd_filepath] = filepath;
@@ -1539,14 +1475,13 @@ bool backup_start(CorruptedPages &corrupted_pages)
 		if (!write_galera_info(mysql_connection)) {
 			return(false);
 		}
+		write_current_binlog_file(mysql_connection);
 	}
 
-	bool with_binlogs = opt_binlog_info == BINLOG_INFO_ON;
+	if (opt_binlog_info == BINLOG_INFO_ON) {
 
-	if (with_binlogs || opt_galera_info) {
-		if (!write_current_binlog_file(mysql_connection, with_binlogs)) {
-			return(false);
-		}
+		lock_binlog_maybe(mysql_connection);
+		write_binlog_info(mysql_connection);
 	}
 
 	if (have_flush_engine_logs && !opt_no_lock) {
@@ -1580,34 +1515,15 @@ void backup_release()
 	}
 }
 
-static const char *default_buffer_pool_file = "ib_buffer_pool";
-
-static
-const char * get_buffer_pool_filename(size_t *length)
-{
-	/* If mariabackup is run for Galera, then the file
-	name is changed to the default so that the receiving
-	node can find this file and rename it according to its
-	settings, otherwise we keep the original file name: */
-	size_t dir_length = 0;
-	const char *dst_name = default_buffer_pool_file;
-	if (!opt_galera_info) {
-		dir_length = dirname_length(buffer_pool_filename);
-		dst_name = buffer_pool_filename + dir_length;
-	}
-	if (length) {
-		*length=dir_length;
-	}
-	return dst_name;
-}
-
 /** Finish after backup_start() and backup_release() */
 bool backup_finish()
 {
 	/* Copy buffer pool dump or LRU dump */
 	if (!opt_rsync) {
 		if (buffer_pool_filename && file_exists(buffer_pool_filename)) {
-			const char *dst_name = get_buffer_pool_filename(NULL);
+			const char *dst_name;
+
+			dst_name = trim_dotslash(buffer_pool_filename);
 			copy_file(ds_data, buffer_pool_filename, dst_name, 0);
 		}
 		if (file_exists("ib_lru_dump")) {
@@ -1696,14 +1612,17 @@ ibx_copy_incremental_over_full()
 
 		/* copy buffer pool dump */
 		if (innobase_buffer_pool_filename) {
-			const char *src_name = get_buffer_pool_filename(NULL);
+			const char *src_name;
+
+			src_name = trim_dotslash(innobase_buffer_pool_filename);
 
 			snprintf(path, sizeof(path), "%s/%s",
 				xtrabackup_incremental_dir,
 				src_name);
 
 			if (file_exists(path)) {
-				copy_file(ds_data, path, src_name, 0);
+				copy_file(ds_data, path,
+					  innobase_buffer_pool_filename, 0);
 			}
 		}
 
@@ -1802,6 +1721,7 @@ apply_log_finish()
 	return(true);
 }
 
+
 bool
 copy_back()
 {
@@ -1822,6 +1742,19 @@ copy_back()
 			return(false);
 		}
 	}
+
+#ifdef _WIN32
+	/* Initialize security descriptor for the new directories
+	to be the same as for datadir */
+	DWORD res = GetNamedSecurityInfoA(mysql_data_home,
+		SE_FILE_OBJECT, DACL_SECURITY_INFORMATION,
+		NULL, NULL, NULL, NULL,
+		&my_dir_security_attributes.lpSecurityDescriptor);
+	if (res != ERROR_SUCCESS) {
+		msg("Unable to read security descriptor of %s",mysql_data_home);
+	}
+#endif
+
 	if (srv_undo_dir && *srv_undo_dir
 		&& !directory_exists(srv_undo_dir, true)) {
 			return(false);
@@ -1856,8 +1789,6 @@ copy_back()
 	}
 
 	srv_max_n_threads = 1000;
-	sync_check_init();
-	ut_crc32_init();
 
 	/* copy undo tablespaces */
 
@@ -1887,24 +1818,25 @@ copy_back()
 	dst_dir = (srv_log_group_home_dir && *srv_log_group_home_dir)
 		? srv_log_group_home_dir : mysql_data_home;
 
-	/* --backup generates a single ib_logfile0, which we must copy
+	/* --backup generates a single LOG_FILE_NAME, which we must copy
 	if it exists. */
 
 	ds_data = ds_create(dst_dir, DS_TYPE_LOCAL);
 	MY_STAT stat_arg;
-	if (!my_stat("ib_logfile0", &stat_arg, MYF(0)) || !stat_arg.st_size) {
+	if (!my_stat(LOG_FILE_NAME, &stat_arg, MYF(0)) || !stat_arg.st_size) {
 		/* After completed --prepare, redo log files are redundant.
 		We must delete any redo logs at the destination, so that
 		the database will not jump to a different log sequence number
 		(LSN). */
 
-		for (uint i = 0; i <= SRV_N_LOG_FILES_MAX + 1; i++) {
-			char filename[FN_REFLEN];
-			snprintf(filename, sizeof filename, "%s/ib_logfile%u",
-				 dst_dir, i);
-			unlink(filename);
-		}
-	} else if (!(ret = copy_or_move_file("ib_logfile0", "ib_logfile0",
+		char filename[FN_REFLEN];
+		snprintf(filename, sizeof filename, "%s/%s0", dst_dir,
+			 LOG_FILE_NAME_PREFIX);
+		unlink(filename);
+		snprintf(filename, sizeof filename, "%s/%s101", dst_dir,
+			 LOG_FILE_NAME_PREFIX);
+		unlink(filename);
+	} else if (!(ret = copy_or_move_file(LOG_FILE_NAME, LOG_FILE_NAME,
 					     dst_dir, 1))) {
 		goto cleanup;
 	}
@@ -1921,9 +1853,8 @@ copy_back()
 	     end(srv_sys_space.end());
 	     iter != end;
 	     ++iter) {
-		const char *filename = base_name(iter->name());
-
-		if (!(ret = copy_or_move_file(filename, iter->name(),
+		const char *filepath = iter->filepath();
+		if (!(ret = copy_or_move_file(base_name(filepath), filepath,
 					      dst_dir, 1))) {
 			goto cleanup;
 		}
@@ -1938,14 +1869,6 @@ copy_back()
 
 	datadir_node_init(&node);
 
-	/* If mariabackup is run for Galera, then the file
-	name is changed to the default so that the receiving
-	node can find this file and rename it according to its
-	settings, otherwise we keep the original file name: */
-	size_t dir_length;
-	const char *src_buffer_pool;
-	src_buffer_pool = get_buffer_pool_filename(&dir_length);
-
 	while (datadir_iter_next(it, &node)) {
 		const char *ext_list[] = {"backup-my.cnf",
 			"xtrabackup_binary", "xtrabackup_binlog_info",
@@ -1954,7 +1877,6 @@ copy_back()
 		const char *filename;
 		char c_tmp;
 		int i_tmp;
-		bool is_ibdata_file;
 
 		if (strstr(node.filepath,"/" ROCKSDB_BACKUP_DIR "/")
 #ifdef _WIN32
@@ -2004,49 +1926,45 @@ copy_back()
 		}
 
 		/* skip the redo log (it was already copied) */
-		if (!strcmp(filename, "ib_logfile0")) {
-			continue;
-		}
-
-		/* skip buffer pool dump */
-		if (!strcmp(filename, src_buffer_pool)) {
+		if (!strcmp(filename, LOG_FILE_NAME)) {
 			continue;
 		}
 
 		/* skip innodb data files */
-		is_ibdata_file = false;
 		for (Tablespace::const_iterator iter(srv_sys_space.begin()),
 		       end(srv_sys_space.end()); iter != end; ++iter) {
-			const char *ibfile = base_name(iter->name());
-			if (strcmp(ibfile, filename) == 0) {
-				is_ibdata_file = true;
-				break;
+			if (!strcmp(base_name(iter->filepath()), filename)) {
+				goto next_file;
 			}
 		}
-		if (is_ibdata_file) {
-			continue;
-		}
 
 		if (!(ret = copy_or_move_file(node.filepath, node.filepath_rel,
 					      mysql_data_home, 1))) {
 			goto cleanup;
 		}
+	next_file:
+		continue;
 	}
 
 	/* copy buffer pool dump */
 
-	if (file_exists(src_buffer_pool)) {
-		char dst_dir[FN_REFLEN];
-		while (IS_TRAILING_SLASH(buffer_pool_filename, dir_length)) {
-			dir_length--;
-		}
-		memcpy(dst_dir, buffer_pool_filename, dir_length);
-		dst_dir[dir_length] = 0;
-		if (!(ret = copy_or_move_file(src_buffer_pool,
-				src_buffer_pool,
-				dst_dir, 1)))
-		{
-			goto cleanup;
+	if (innobase_buffer_pool_filename) {
+		const char *src_name;
+		char path[FN_REFLEN];
+
+		src_name = trim_dotslash(innobase_buffer_pool_filename);
+
+		snprintf(path, sizeof(path), "%s/%s",
+			mysql_data_home,
+			src_name);
+
+		/* could be already copied with other files
+		from data directory */
+		if (file_exists(src_name) &&
+			!file_exists(innobase_buffer_pool_filename)) {
+			copy_or_move_file(src_name,
+					  innobase_buffer_pool_filename,
+					  mysql_data_home, 0);
 		}
 	}
 
@@ -2065,7 +1983,6 @@ cleanup:
 
 	ds_data = NULL;
 
-	sync_check_close();
 	return(ret);
 }
 
@@ -2113,13 +2030,10 @@ decrypt_decompress_file(const char *filepath, uint thread_n)
  	return(true);
 }
 
-static
-os_thread_ret_t STDCALL
-decrypt_decompress_thread_func(void *arg)
+static void decrypt_decompress_thread_func(datadir_thread_ctxt_t *ctxt)
 {
 	bool ret = true;
 	datadir_node_t node;
-	datadir_thread_ctxt_t *ctxt = (datadir_thread_ctxt_t *)(arg);
 
 	datadir_node_init(&node);
 
@@ -2149,9 +2063,6 @@ cleanup:
 	pthread_mutex_unlock(ctxt->count_mutex);
 
 	ctxt->ret = ret;
-
-	os_thread_exit();
-	OS_THREAD_DUMMY_RETURN;
 }
 
 bool
@@ -2161,7 +2072,6 @@ decrypt_decompress()
 	datadir_iter_t *it = NULL;
 
 	srv_max_n_threads = 1000;
-	sync_check_init();
 
 	/* cd to backup directory */
 	if (my_setwd(xtrabackup_target_dir, MYF(MY_WME)))
@@ -2190,8 +2100,6 @@ decrypt_decompress()
 
 	ds_data = NULL;
 
-	sync_check_close();
-
 	return(ret);
 }
 
@@ -2212,7 +2120,14 @@ static bool backup_files_from_datadir(const char *dir_path)
 		if (info.type != OS_FILE_TYPE_FILE)
 			continue;
 
-		const char *pname = strrchr(info.name, OS_PATH_SEPARATOR);
+		const char *pname = strrchr(info.name, '/');
+#ifdef _WIN32
+		if (const char *last = strrchr(info.name, '\\')) {
+			if (!pname || last >pname) {
+				pname = last;
+			}
+		}
+#endif
 		if (!pname)
 			pname = info.name;
 
@@ -2229,7 +2144,7 @@ static bool backup_files_from_datadir(const char *dir_path)
 			unlink(info.name);
 
 		std::string full_path(dir_path);
-		full_path.append(1, OS_PATH_SEPARATOR).append(info.name);
+		full_path.append(1, '/').append(info.name);
 		if (!(ret = copy_file(ds_data, full_path.c_str() , info.name, 1)))
 			break;
 	}
