@@ -19,7 +19,16 @@
  * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1335, USA
  */
 
-#include "caam_driver.h"
+#ifdef HAVE_CONFIG_H
+    #include <config.h>
+#endif
+
+#include <wolfssl/wolfcrypt/settings.h>
+
+#if defined(__QNX__) || defined(__QNXNTO__)
+
+#include <wolfssl/wolfcrypt/port/caam/caam_driver.h>
+#include <wolfssl/version.h>
 
 #include <errno.h>
 #include <stdio.h>
@@ -192,7 +201,7 @@ void CAAM_ADR_UNMAP(void* vaddr, unsigned int out, int outSz,
 }
 
 
-/* syncoronize virtual buffer with physical
+/* synchronize virtual buffer with physical
  * return 0 on success */
 int CAAM_ADR_SYNC(void* vaddr, int sz)
 {
@@ -1114,7 +1123,7 @@ int io_devctl (resmgr_context_t *ctp, io_devctl_t *msg, iofunc_ocb_t *ocb)
             break;
 
         case WC_CAAM_FIND_PART:
-            ret = caamFindUnusuedPartition();
+            ret = caamFindUnusedPartition();
             if (ret < 0) {
                 /* none found, try again later */
                 return EAGAIN;
@@ -1190,8 +1199,9 @@ static int getSupported(char* in)
 #endif
 
 char cannedResponse[] = {
-        "wolfCrypt QNX CAAM driver version 4.8.1\n"
-        "Supports:\n"
+        "wolfCrypt QNX CAAM driver version "
+        LIBWOLFSSL_VERSION_STRING
+        "\nSupports:\n"
         "\tAES-CMAC\n"
         "\tECC (sign, verify, ecdh, keygen)\n"
         "\tBlobs (black and red)\n"
@@ -1287,3 +1297,4 @@ int main(int argc, char *argv[])
     return 0;
 }
 
+#endif /* __QNX__ || __QNXNTO__ */
