@@ -1,5 +1,5 @@
 /* Copyright (c) 2009, 2013, Oracle and/or its affiliates.
-   Copyright (c) 2013, 2020, MariaDB
+   Copyright (c) 2009, 2020, MariaDB Corporation.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -16,7 +16,7 @@
 
 /* see include/mysql/service_debug_sync.h for debug sync documentation */
 
-#include <my_global.h>
+#include "mariadb.h"
 #include "debug_sync.h"
 
 #if defined(ENABLED_DEBUG_SYNC)
@@ -121,6 +121,25 @@ static void init_debug_sync_psi_keys(void)
 #endif /* HAVE_PSI_INTERFACE */
 
 
+/**
+  Set the THD::proc_info without instrumentation.
+  This method is private to DEBUG_SYNC,
+  and on purpose avoid any use of:
+  - the SHOW PROFILE instrumentation
+  - the PERFORMANCE_SCHEMA instrumentation
+  so that using DEBUG_SYNC() in the server code
+  does not cause the instrumentations to record
+  spurious data.
+*/
+static const char*
+debug_sync_thd_proc_info(THD *thd, const char* info)
+{
+  const char* old_proc_info= thd->proc_info;
+  thd->proc_info= info;
+  return old_proc_info;
+}
+
+
 /**
   Initialize the debug sync facility at server start.
 
@@ -182,13 +201,12 @@ void debug_sync_end(void)
 
     /* Print statistics. */
     {
-      char llbuff[22];
-      sql_print_information("Debug sync points hit:                   %22s",
-                            llstr(debug_sync_global.dsp_hits, llbuff));
-      sql_print_information("Debug sync points executed:              %22s",
-                            llstr(debug_sync_global.dsp_executed, llbuff));
-      sql_print_information("Debug sync points max active per thread: %22s",
-                            llstr(debug_sync_global.dsp_max_active, llbuff));
+      sql_print_information("Debug sync points hit:                   %lld",
+                            debug_sync_global.dsp_hits);
+      sql_print_information("Debug sync points executed:              %lld",
+                            debug_sync_global.dsp_executed);
+      sql_print_information("Debug sync points max active per thread: %lld",
+                            debug_sync_global.dsp_max_active);
     }
   }
 
@@ -236,7 +254,8 @@ void debug_sync_init_thread(THD *thd)
   if (opt_debug_sync_timeout)
   {
     thd->debug_sync_control= (st_debug_sync_control*)
-      my_malloc(sizeof(st_debug_sync_control),
+      my_malloc(PSI_NOT_INSTRUMENTED,
+                sizeof(st_debug_sync_control),
                 MYF(MY_WME | MY_ZEROFILL | MY_THREAD_SPECIFIC));
     if (!thd->debug_sync_control)
     {
@@ -267,12 +286,6 @@ void debug_sync_end_thread(THD *thd)
   {
     st_debug_sync_control *ds_control= thd->debug_sync_control;
 
-    /*
-      This synchronization point can be used to synchronize on thread end.
-      This is the latest point in a THD's life, where this can be done.
-    */
-    DEBUG_SYNC(thd, "thread_end");
-
     if (ds_control->ds_action)
     {
       st_debug_sync_action *action= ds_control->ds_action;
@@ -302,6 +315,20 @@ void debug_sync_end_thread(THD *thd)
 }
 
 
+void debug_sync_reset_thread(THD *thd)
+{
+  if (thd->debug_sync_control)
+  {
+    /*
+      This synchronization point can be used to synchronize on thread end.
+      This is the latest point in a THD's life, where this can be done.
+    */
+    DEBUG_SYNC(thd, "thread_end");
+    thd->debug_sync_control->ds_active= 0;
+  }
+}
+
+
 /**
   Move a string by length.
 
@@ -467,13 +494,13 @@ static int debug_sync_qsort_cmp(const void* arg1, const void* arg2)
 static st_debug_sync_action *debug_sync_find(st_debug_sync_action *actionarr,
                                              int quantity,
                                              const char *dsp_name,
-                                             uint name_len)
+                                             size_t name_len)
 {
   st_debug_sync_action  *action;
   int                   low ;
   int                   high ;
   int                   mid ;
-  int                   diff ;
+  ssize_t               diff ;
   DBUG_ASSERT(actionarr);
   DBUG_ASSERT(dsp_name);
   DBUG_ASSERT(name_len);
@@ -658,7 +685,7 @@ static st_debug_sync_action *debug_sync_get_action(THD *thd,
     if (ds_control->ds_active > ds_control->ds_allocated)
     {
       uint new_alloc= ds_control->ds_active + 3;
-      void *new_action= my_realloc(ds_control->ds_action,
+      void *new_action= my_realloc(PSI_NOT_INSTRUMENTED, ds_control->ds_action,
                                    new_alloc * sizeof(st_debug_sync_action),
                                    MYF(MY_WME | MY_ALLOW_ZERO_PTR));
       if (!new_action)
@@ -792,7 +819,7 @@ static bool debug_sync_set_action(THD *thd, st_debug_sync_action *action)
       and shall not be reported as a result of SET DEBUG_SYNC.
       Hence, we check for the first condition above.
     */
-    if (thd->is_error())
+    if (unlikely(thd->is_error()))
       DBUG_RETURN(TRUE);
   }
 
@@ -857,8 +884,7 @@ static char *debug_sync_token(char **token_p, uint *token_length_p,
   DBUG_ASSERT(ptr);
 
   /* Skip leading space */
-  ptr+= system_charset_info->cset->scan(system_charset_info,
-                                        ptr, ptrend, MY_SEQ_SPACES);
+  ptr+= system_charset_info->scan(ptr, ptrend, MY_SEQ_SPACES);
   if (!*ptr)
   {
     ptr= NULL;
@@ -869,8 +895,7 @@ static char *debug_sync_token(char **token_p, uint *token_length_p,
   *token_p= ptr;
 
   /* Find token end. */
-  ptr+= system_charset_info->cset->scan(system_charset_info,
-                                        ptr, ptrend, MY_SEQ_NONSPACES);
+  ptr+= system_charset_info->scan(ptr, ptrend, MY_SEQ_NONSPACES);
 
   /* Get token length. */
   *token_length_p= (uint)(ptr - *token_p);
@@ -880,7 +905,7 @@ static char *debug_sync_token(char **token_p, uint *token_length_p,
   {
     DBUG_ASSERT(ptr < ptrend);
     /* Get terminator character length. */
-    uint mbspacelen= my_charlen_fix(system_charset_info, ptr, ptrend);
+    uint mbspacelen= system_charset_info->charlen_fix(ptr, ptrend);
 
     /* Terminate token. */
     *ptr= '\0';
@@ -889,8 +914,7 @@ static char *debug_sync_token(char **token_p, uint *token_length_p,
     ptr+= mbspacelen;
 
     /* Skip trailing space */
-    ptr+= system_charset_info->cset->scan(system_charset_info,
-                                          ptr, ptrend, MY_SEQ_SPACES);
+    ptr+= system_charset_info->scan(ptr, ptrend, MY_SEQ_SPACES);
   }
 
  end:
@@ -971,6 +995,15 @@ static char *debug_sync_number(ulong *number_p, char *actstrptr,
     The input string needs to be ASCII NUL ('\0') terminated. We split
     nul-terminated tokens in it without copy.
 
+  @note
+    The current implementation does not support two 'now SIGNAL xxx' commands
+    in a row for multiple threads as the first one can get lost while
+    the waiting threads are sleeping on mysql_cond_timedwait().
+    One reason for this is that the signal name is stored in a global variable
+    that is overwritten.  A better way would be to store all signals in
+    an array together with a 'time' when the signal was sent. This array
+    should be checked on broadcast.
+
   @see the function comment of debug_sync_token() for more constraints
     for the string.
 */
@@ -1368,7 +1401,7 @@ static void debug_sync_execute(THD *thd, st_debug_sync_action *action)
       strxnmov(ds_control->ds_proc_info, sizeof(ds_control->ds_proc_info)-1,
                "debug sync point: ", action->sync_point.c_ptr(), NullS);
       old_proc_info= thd->proc_info;
-      thd_proc_info(thd, ds_control->ds_proc_info);
+      debug_sync_thd_proc_info(thd, ds_control->ds_proc_info);
     }
 
     /*
@@ -1450,16 +1483,14 @@ static void debug_sync_execute(THD *thd, st_debug_sync_action *action)
             DBUG_PRINT("debug_sync",
                        ("awoke from %s  global: %s  error: %d",
                         sig_wait, sig_glob, error));});
-        if (error == ETIMEDOUT || error == ETIME)
+        if (unlikely(error == ETIMEDOUT || error == ETIME))
         {
           // We should not make the statement fail, even if in strict mode.
-          const bool save_abort_on_warning= thd->abort_on_warning;
-          thd->abort_on_warning= false;
+          Abort_on_warning_instant_set aws(thd, false);
           push_warning(thd, Sql_condition::WARN_LEVEL_WARN,
                        ER_DEBUG_SYNC_TIMEOUT,
                        ER_THD(thd, ER_DEBUG_SYNC_TIMEOUT));
-          thd->abort_on_warning= save_abort_on_warning;
-          DBUG_EXECUTE_IF("debug_sync_abort_on_timeout", DBUG_ABORT(););
+          DBUG_EXECUTE_IF("debug_sync_abort_on_timeout", DBUG_ASSERT(0););
           break;
         }
         error= 0;
@@ -1488,11 +1519,11 @@ static void debug_sync_execute(THD *thd, st_debug_sync_action *action)
         mysql_mutex_lock(&thd->mysys_var->mutex);
         thd->mysys_var->current_mutex= old_mutex;
         thd->mysys_var->current_cond= old_cond;
-        thd_proc_info(thd, old_proc_info);
+        debug_sync_thd_proc_info(thd, old_proc_info);
         mysql_mutex_unlock(&thd->mysys_var->mutex);
       }
       else
-        thd_proc_info(thd, old_proc_info);
+        debug_sync_thd_proc_info(thd, old_proc_info);
     }
     else
     {
