@@ -240,7 +240,8 @@ public:
     threads, the SQL thread sets it to unblock the I/O thread and make it
     temporarily forget about the constraint.
   */
-  ulonglong log_space_limit,log_space_total;
+  ulonglong log_space_limit;
+  Atomic_counter<uint64> log_space_total;
   bool ignore_log_space_limit;
 
   /*
@@ -273,7 +274,7 @@ public:
     Needed for problems when slave stops and we want to restart it
     skipping one or more events in the master log that have caused
     errors, and have been manually applied by DBA already.
-    Must be ulong as it's refered to from set_var.cc
+    Must be ulong as it's referred to from set_var.cc
   */
   volatile ulonglong slave_skip_counter;
   ulonglong max_relay_log_size;
@@ -347,7 +348,7 @@ public:
     Number of executed events for SLAVE STATUS.
     Protected by slave_executed_entries_lock
   */
-  int64 executed_entries;
+  Atomic_counter<uint32_t> executed_entries;
 
   /*
     If the end of the hot relay log is made of master's events ignored by the
@@ -387,11 +388,11 @@ public:
   */
   slave_connection_state restart_gtid_pos;
 
-  Relay_log_info(bool is_slave_recovery);
+  Relay_log_info(bool is_slave_recovery, const char* thread_name= "SQL");
   ~Relay_log_info();
 
   /*
-    Invalidate cached until_log_name and group_relay_log_name comparison 
+    Invalidate cached until_log_name and group_relay_log_name comparison
     result. Should be called after any update of group_realy_log_name if
     there chances that sql_thread is running.
   */
@@ -538,7 +539,8 @@ public:
   int32 get_sql_delay() { return sql_delay; }
   void set_sql_delay(int32 _sql_delay) { sql_delay= _sql_delay; }
   time_t get_sql_delay_end() { return sql_delay_end; }
-
+  rpl_gtid last_seen_gtid;
+  ulong last_trans_retry_count;
 private:
 
 
@@ -623,10 +625,20 @@ struct inuse_relaylog {
   /* Number of events in this relay log queued for worker threads. */
   int64 queued_count;
   /* Number of events completed by worker threads. */
-  volatile int64 dequeued_count;
+  Atomic_counter<int64> dequeued_count;
   /* Set when all events have been read from a relaylog. */
   bool completed;
   char name[FN_REFLEN];
+
+  inuse_relaylog(Relay_log_info *rli_arg, rpl_gtid *relay_log_state_arg,
+                 uint32 relay_log_state_count_arg,
+                 const char *name_arg):
+    next(0), rli(rli_arg), relay_log_state(relay_log_state_arg),
+    relay_log_state_count(relay_log_state_count_arg), queued_count(0),
+    dequeued_count(0), completed(false)
+  {
+    strmake_buf(name, name_arg);
+  }
 };
 
 
@@ -767,16 +779,11 @@ struct rpl_group_info
     Runtime state for printing a note when slave is taking
     too long while processing a row event.
    */
-  time_t row_stmt_start_timestamp;
+  longlong row_stmt_start_timestamp;
   bool long_find_row_note_printed;
   /* Needs room for "Gtid D-S-N\x00". */
   char gtid_info_buf[5+10+1+10+1+20+1];
 
-  /* List of not yet committed deletions in mysql.gtid_slave_pos. */
-  rpl_slave_state::list_element *pending_gtid_delete_list;
-  /* Domain associated with pending_gtid_delete_list. */
-  uint32 pending_gtid_delete_list_domain;
-
   /*
     The timestamp, from the master, of the commit event.
     Used to do delayed update of rli->last_master_timestamp, for getting
@@ -918,23 +925,15 @@ struct rpl_group_info
   char *gtid_info();
   void unmark_start_commit();
 
-  static void pending_gtid_deletes_free(rpl_slave_state::list_element *list);
-  void pending_gtid_deletes_save(uint32 domain_id,
-                                 rpl_slave_state::list_element *list);
-  void pending_gtid_deletes_put_back();
-  void pending_gtid_deletes_clear();
-
-  time_t get_row_stmt_start_timestamp()
+  longlong get_row_stmt_start_timestamp()
   {
     return row_stmt_start_timestamp;
   }
 
-  time_t set_row_stmt_start_timestamp()
+  void set_row_stmt_start_timestamp()
   {
     if (row_stmt_start_timestamp == 0)
-      row_stmt_start_timestamp= my_time(0);
-
-    return row_stmt_start_timestamp;
+      row_stmt_start_timestamp= microsecond_interval_timer();
   }
 
   void reset_row_stmt_start_timestamp()
@@ -993,6 +992,7 @@ extern struct rpl_slave_state *rpl_global_gtid_slave_state;
 extern gtid_waiting rpl_global_gtid_waiting;
 
 int rpl_load_gtid_slave_state(THD *thd);
+int find_gtid_slave_pos_tables(THD *thd);
 int event_group_new_gtid(rpl_group_info *rgi, Gtid_log_event *gev);
 void delete_or_keep_event_post_apply(rpl_group_info *rgi,
                                      Log_event_type typ, Log_event *ev);
