@@ -74,7 +74,7 @@ class Json_writer;
 *************************************************************************************/
 
 
-const int FAKE_SELECT_LEX_ID= (int)UINT_MAX;
+const uint FAKE_SELECT_LEX_ID= UINT_MAX;
 
 class Explain_query;
 
@@ -108,7 +108,7 @@ public:
   };
 
   virtual enum explain_node_type get_type()= 0;
-  virtual int get_select_id()= 0;
+  virtual uint get_select_id()= 0;
 
   /**
     expression cache statistics
@@ -168,9 +168,9 @@ public:
 
   bool add_table(Explain_table_access *tab, Explain_query *query);
 
-  int get_select_id() { return select_id; }
+  uint get_select_id() { return select_id; }
 
-  int select_id;
+  uint select_id;
 
   int print_explain(Explain_query *query, select_result_sink *output,
                     uint8 explain_flags, bool is_analyze);
@@ -212,6 +212,8 @@ public:
 #ifndef DBUG_OFF
     select_lex(NULL),
 #endif
+    linkage(UNSPECIFIED_TYPE),
+    is_lateral(false),
     message(NULL),
     having(NULL), having_value(Item::COND_UNDEF),
     using_temporary(false), using_filesort(false),
@@ -219,11 +221,15 @@ public:
     aggr_tree(NULL)
   {}
 
+  void add_linkage(Json_writer *writer);
+
 public:
 #ifndef DBUG_OFF
   SELECT_LEX *select_lex;
 #endif
   const char *select_type;
+  enum sub_select_type linkage;
+  bool is_lateral;
 
   /*
     If message != NULL, this is a degenerate join plan, and all subsequent
@@ -326,7 +332,11 @@ public:
 
 /////////////////////////////////////////////////////////////////////////////
 
-/* 
+extern const char *unit_operation_text[4];
+extern const char *pushed_derived_text;
+extern const char *pushed_select_text;
+
+/*
   Explain structure for a UNION.
 
   A UNION may or may not have "Using filesort".
@@ -336,14 +346,15 @@ class Explain_union : public Explain_node
 {
 public:
   Explain_union(MEM_ROOT *root, bool is_analyze) : 
-    Explain_node(root),
+    Explain_node(root), union_members(PSI_INSTRUMENT_MEM),
     is_recursive_cte(false),
     fake_select_lex_explain(root, is_analyze)
   {}
 
   enum explain_node_type get_type() { return EXPLAIN_UNION; }
+  unit_common_op operation;
 
-  int get_select_id()
+  uint get_select_id()
   {
     DBUG_ASSERT(union_members.elements() > 0);
     return union_members.at(0);
@@ -541,6 +552,7 @@ enum explain_extra_tag
   ET_CONST_ROW_NOT_FOUND,
   ET_UNIQUE_ROW_NOT_FOUND,
   ET_IMPOSSIBLE_ON_CONDITION,
+  ET_TABLE_FUNCTION,
 
   ET_total
 };
@@ -579,6 +591,8 @@ class Explain_index_use : public Sql_alloc
 {
   char *key_name;
   uint key_len;
+  char *filter_name;
+  uint filter_len;
 public:
   String_list key_parts_list;
   
@@ -591,12 +605,46 @@ public:
   {
     key_name= NULL;
     key_len= (uint)-1;
+    filter_name= NULL;
+    filter_len= (uint)-1;
   }
-  void set(MEM_ROOT *root, KEY *key_name, uint key_len_arg);
-  void set_pseudo_key(MEM_ROOT *root, const char *key_name);
+  bool set(MEM_ROOT *root, KEY *key_name, uint key_len_arg);
+  bool set_pseudo_key(MEM_ROOT *root, const char *key_name);
 
   inline const char *get_key_name() const { return key_name; }
   inline uint get_key_len() const { return key_len; }
+  //inline const char *get_filter_name() const { return filter_name; }
+};
+
+
+/*
+  Query Plan data structure for Rowid filter.
+*/
+class Explain_rowid_filter : public Sql_alloc
+{
+public:
+  /* Quick select used to collect the rowids into filter */
+  Explain_quick_select *quick;
+
+  /* How many rows the above quick select is expected to return */
+  ha_rows rows;
+
+  /* Expected selectivity for the filter */
+  double selectivity;
+
+  /* Tracker with the information about how rowid filter is executed */
+  Rowid_filter_tracker *tracker;
+
+  void print_explain_json(Explain_query *query, Json_writer *writer,
+                          bool is_analyze);
+
+  /*
+    TODO:
+      Here should be ANALYZE members:
+      - r_rows for the quick select
+      - An object that tracked the table access time
+      - real selectivity of the filter.
+  */
 };
 
 
@@ -666,6 +714,7 @@ public:
   void print_json(Json_writer *writer, bool is_analyze);
 };
 
+
 /*
   EXPLAIN data structure for a single JOIN_TAB.
 */
@@ -685,7 +734,8 @@ public:
     cache_cond(NULL),
     pushed_index_cond(NULL),
     sjm_nest(NULL),
-    pre_join_sort(NULL)
+    pre_join_sort(NULL),
+    rowid_filter(NULL)
   {}
   ~Explain_table_access() { delete sjm_nest; }
 
@@ -790,8 +840,12 @@ public:
   /* Tracker for reading the table */
   Table_access_tracker tracker;
   Exec_time_tracker op_tracker;
+  Gap_time_tracker extra_time_tracker;
+
   Table_access_tracker jbuf_tracker;
   
+  Explain_rowid_filter *rowid_filter;
+
   int print_explain(select_result_sink *output, uint8 explain_flags, 
                     bool is_analyze,
                     uint select_id, const char *select_type,
@@ -802,7 +856,7 @@ public:
 private:
   void append_tag_name(String *str, enum explain_extra_tag tag);
   void fill_key_str(String *key_str, bool is_json) const;
-  void fill_key_len_str(String *key_len_str) const;
+  void fill_key_len_str(String *key_len_str, bool is_json) const;
   double get_r_filtered();
   void tag_to_json(Json_writer *writer, enum explain_extra_tag tag);
 };
@@ -828,7 +882,7 @@ public:
   {}
 
   virtual enum explain_node_type get_type() { return EXPLAIN_UPDATE; }
-  virtual int get_select_id() { return 1; /* always root */ }
+  virtual uint get_select_id() { return 1; /* always root */ }
 
   const char *select_type;
 
@@ -908,7 +962,7 @@ public:
   StringBuffer<64> table_name;
 
   enum explain_node_type get_type() { return EXPLAIN_INSERT; }
-  int get_select_id() { return 1; /* always root */ }
+  uint get_select_id() { return 1; /* always root */ }
 
   int print_explain(Explain_query *query, select_result_sink *output, 
                     uint8 explain_flags, bool is_analyze);
@@ -935,7 +989,7 @@ public:
   bool deleting_all_rows;
 
   virtual enum explain_node_type get_type() { return EXPLAIN_DELETE; }
-  virtual int get_select_id() { return 1; /* always root */ }
+  virtual uint get_select_id() { return 1; /* always root */ }
 
   virtual int print_explain(Explain_query *query, select_result_sink *output, 
                             uint8 explain_flags, bool is_analyze);
