@@ -1,7 +1,7 @@
 /*****************************************************************************
 
 Copyright (c) 2000, 2017, Oracle and/or its affiliates. All Rights Reserved.
-Copyright (c) 2013, 2022, MariaDB Corporation.
+Copyright (c) 2013, 2021, MariaDB Corporation.
 
 This program is free software; you can redistribute it and/or modify it under
 the terms of the GNU General Public License as published by the Free Software
@@ -16,10 +16,9 @@ this program; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1335 USA
 
 *****************************************************************************/
-
 #ifdef WITH_WSREP
-# include <mysql/service_wsrep.h>
-# include "../../../wsrep/wsrep_api.h"
+#include "wsrep_api.h"
+#include <mysql/service_wsrep.h>
 #endif /* WITH_WSREP */
 
 #include "table.h"
@@ -52,42 +51,39 @@ struct ha_table_option_struct
 	uint		encryption;		/*!<  DEFAULT, ON, OFF */
 	ulonglong	encryption_key_id;	/*!< encryption key id  */
 };
-/* JAN: TODO: MySQL 5.7 handler.h */
-struct st_handler_tablename
-{
-  const char *db;
-  const char *tablename;
-};
+
 /** The class defining a handle to an Innodb table */
-class ha_innobase: public handler
+class ha_innobase final : public handler
 {
 public:
 	ha_innobase(handlerton* hton, TABLE_SHARE* table_arg);
-	~ha_innobase();
+	~ha_innobase() override;
+
+	/** @return the transaction that last modified the table definition
+	@see dict_table_t::def_trx_id */
+	ulonglong table_version() const override;
 
 	/** Get the row type from the storage engine.  If this method returns
 	ROW_TYPE_NOT_USED, the information in HA_CREATE_INFO should be used. */
-	enum row_type get_row_type() const;
+        enum row_type get_row_type() const override;
 
-	const char* table_type() const;
+        const char* table_type() const override;
 
-	const char* index_type(uint key_number);
+	const char* index_type(uint key_number) override;
 
-	const char** bas_ext() const;
+	Table_flags table_flags() const override;
 
-	Table_flags table_flags() const;
+	ulong index_flags(uint idx, uint part, bool all_parts) const override;
 
-	ulong index_flags(uint idx, uint part, bool all_parts) const;
+	uint max_supported_keys() const override;
 
-	uint max_supported_keys() const;
+	uint max_supported_key_length() const override;
 
-	uint max_supported_key_length() const;
+	uint max_supported_key_part_length() const override;
 
-	uint max_supported_key_part_length() const;
+	const key_map* keys_to_use_for_scanning() override;
 
-	const key_map* keys_to_use_for_scanning();
-
-	void column_bitmaps_signal();
+	void column_bitmaps_signal() override;
 
 	/** Opens dictionary table object using table name. For partition, we need to
 	try alternative lower/upper case names to support moving data files across
@@ -103,95 +99,98 @@ public:
 		bool			is_partition,
 		dict_err_ignore_t	ignore_err);
 
-	int open(const char *name, int mode, uint test_if_locked);
+	int open(const char *name, int mode, uint test_if_locked) override;
 
-	handler* clone(const char *name, MEM_ROOT *mem_root);
+	handler* clone(const char *name, MEM_ROOT *mem_root) override;
 
-	int close(void);
+	int close(void) override;
 
-	double scan_time();
+	double scan_time() override;
 
-	double read_time(uint index, uint ranges, ha_rows rows);
+	double read_time(uint index, uint ranges, ha_rows rows) override;
 
-	int write_row(uchar * buf);
+	int delete_all_rows() override;
 
-	int update_row(const uchar * old_data, const uchar * new_data);
+	int write_row(const uchar * buf) override;
 
-	int delete_row(const uchar * buf);
+	int update_row(const uchar * old_data, const uchar * new_data) override;
 
-	bool was_semi_consistent_read();
+	int delete_row(const uchar * buf) override;
 
-	void try_semi_consistent_read(bool yes);
+	bool was_semi_consistent_read() override;
 
-	void unlock_row();
+	void try_semi_consistent_read(bool yes) override;
 
-	int index_init(uint index, bool sorted);
+	void unlock_row() override;
 
-	int index_end();
+	int index_init(uint index, bool sorted) override;
+
+	int index_end() override;
 
 	int index_read(
 		uchar*			buf,
 		const uchar*		key,
 		uint			key_len,
-		ha_rkey_function	find_flag);
+		ha_rkey_function	find_flag) override;
 
-	int index_read_last(uchar * buf, const uchar * key, uint key_len);
+	int index_read_last(uchar * buf, const uchar * key,
+			    uint key_len) override;
 
-	int index_next(uchar * buf);
+        int index_next(uchar * buf) override;
 
-	int index_next_same(uchar * buf, const uchar *key, uint keylen);
+	int index_next_same(uchar * buf, const uchar * key,
+			    uint keylen) override;
 
-	int index_prev(uchar * buf);
+	int index_prev(uchar * buf) override;
 
-	int index_first(uchar * buf);
+	int index_first(uchar * buf) override;
 
-	int index_last(uchar * buf);
+	int index_last(uchar * buf) override;
 
 	/* Copy a cached MySQL row. If requested, also avoids
 	overwriting non-read columns. */
 	void copy_cached_row(uchar *to_rec, const uchar *from_rec,
 				uint rec_length);
-	int rnd_init(bool scan);
+	int rnd_init(bool scan) override;
 
-	int rnd_end();
+	int rnd_end() override;
 
-	int rnd_next(uchar *buf);
+	int rnd_next(uchar *buf) override;
 
-	int rnd_pos(uchar * buf, uchar *pos);
+	int rnd_pos(uchar * buf, uchar *pos) override;
 
-	int ft_init();
-	void ft_end() { rnd_end(); }
-	FT_INFO *ft_init_ext(uint flags, uint inx, String* key);
-	int ft_read(uchar* buf);
+	int ft_init() override;
+	void ft_end() override { rnd_end(); }
+	FT_INFO *ft_init_ext(uint flags, uint inx, String* key) override;
+	int ft_read(uchar* buf) override;
 
-	void position(const uchar *record);
+	void position(const uchar *record) override;
 
-	int info(uint);
+	int info(uint) override;
 
-	int analyze(THD* thd,HA_CHECK_OPT* check_opt);
+	int analyze(THD* thd,HA_CHECK_OPT* check_opt) override;
 
-	int optimize(THD* thd,HA_CHECK_OPT* check_opt);
+	int optimize(THD* thd,HA_CHECK_OPT* check_opt) override;
 
-	int discard_or_import_tablespace(my_bool discard);
+	int discard_or_import_tablespace(my_bool discard) override;
 
-	int extra(ha_extra_function operation);
+	int extra(ha_extra_function operation) override;
 
-	int reset();
+	int reset() override;
 
-	int external_lock(THD *thd, int lock_type);
+	int external_lock(THD *thd, int lock_type) override;
 
-	int start_stmt(THD *thd, thr_lock_type lock_type);
-
-	void position(uchar *record);
+	int start_stmt(THD *thd, thr_lock_type lock_type) override;
 
 	ha_rows records_in_range(
-		uint			inx,
-		key_range*		min_key,
-		key_range*		max_key);
+                uint                    inx,
+                const key_range*        min_key,
+                const key_range*        max_key,
+                page_range*             pages) override;
 
-	ha_rows estimate_rows_upper_bound();
+	ha_rows estimate_rows_upper_bound() override;
 
-	void update_create_info(HA_CREATE_INFO* create_info);
+	void update_create_info(HA_CREATE_INFO* create_info) override;
 
 	inline int create(
 		const char*		name,
@@ -203,60 +202,55 @@ public:
 	int create(
 		const char*		name,
 		TABLE*			form,
-		HA_CREATE_INFO*		create_info);
+		HA_CREATE_INFO*		create_info) override;
 
-	const char* check_table_options(THD *thd, TABLE* table,
-		HA_CREATE_INFO*	create_info, const bool use_tablespace, const ulint file_format);
+	int truncate() override;
 
-	inline int delete_table(const char* name, enum_sql_command sqlcom);
+	int delete_table(const char *name) override;
 
-	int truncate();
+	int rename_table(const char* from, const char* to) override;
+	inline int defragment_table();
+	int check(THD* thd, HA_CHECK_OPT* check_opt) override;
 
-	int delete_table(const char *name);
+	inline void reload_statistics();
 
-	int rename_table(const char* from, const char* to);
-	inline int defragment_table(const char* name);
-	int check(THD* thd, HA_CHECK_OPT* check_opt);
+	char* get_foreign_key_create_info() override;
 
-	char* get_foreign_key_create_info();
-
-	int get_foreign_key_list(THD *thd, List<FOREIGN_KEY_INFO> *f_key_list);
+        int get_foreign_key_list(THD *thd,
+                                 List<FOREIGN_KEY_INFO> *f_key_list) override;
 
 	int get_parent_foreign_key_list(
 		THD*			thd,
-		List<FOREIGN_KEY_INFO>*	f_key_list);
-	int get_cascade_foreign_key_table_list(
-		THD*				thd,
-		List<st_handler_tablename>*	fk_table_list);
+		List<FOREIGN_KEY_INFO>*	f_key_list) override;
 
+	bool can_switch_engines() override;
 
-	bool can_switch_engines();
+	uint referenced_by_foreign_key() override;
 
-	uint referenced_by_foreign_key();
+	void free_foreign_key_create_info(char* str) override;
 
-	void free_foreign_key_create_info(char* str);
-
-	uint lock_count(void) const;
+	uint lock_count(void) const override;
 
 	THR_LOCK_DATA** store_lock(
 		THD*			thd,
 		THR_LOCK_DATA**		to,
-		thr_lock_type		lock_type);
+		thr_lock_type		lock_type) override;
 
-	void init_table_handle_for_HANDLER();
+	void init_table_handle_for_HANDLER() override;
 
-	virtual void get_auto_increment(
+	void get_auto_increment(
 		ulonglong		offset,
 		ulonglong		increment,
 		ulonglong		nb_desired_values,
 		ulonglong*		first_value,
-		ulonglong*		nb_reserved_values);
+		ulonglong*		nb_reserved_values) override;
+	int reset_auto_increment(ulonglong value) override;
 
-	virtual bool get_error_message(int error, String *buf);
+	bool get_error_message(int error, String *buf) override;
 
-	virtual bool get_foreign_dup_key(char*, uint, char*, uint);
+	bool get_foreign_dup_key(char*, uint, char*, uint) override;
 
-	uint8 table_cache_type();
+	uint8 table_cache_type() override;
 
 	/**
 	Ask handler about permission to cache table during query registration
@@ -266,11 +260,9 @@ public:
 		const char*		table_key,
 		uint			key_length,
 		qc_engine_callback*	call_back,
-		ulonglong*		engine_data);
+		ulonglong*		engine_data) override;
 
-	bool primary_key_is_clustered();
-
-	int cmp_ref(const uchar* ref1, const uchar* ref2);
+	int cmp_ref(const uchar* ref1, const uchar* ref2) override;
 
 	/** On-line ALTER TABLE interface @see handler0alter.cc @{ */
 
@@ -300,7 +292,7 @@ public:
 
 	enum_alter_inplace_result check_if_supported_inplace_alter(
 		TABLE*			altered_table,
-		Alter_inplace_info*	ha_alter_info);
+		Alter_inplace_info*	ha_alter_info) override;
 
 	/** Allows InnoDB to update internal structures with concurrent
 	writes blocked (provided that check_if_supported_inplace_alter()
@@ -316,7 +308,7 @@ public:
 	*/
 	bool prepare_inplace_alter_table(
 		TABLE*			altered_table,
-		Alter_inplace_info*	ha_alter_info);
+		Alter_inplace_info*	ha_alter_info) override;
 
 	/** Alter the table structure in-place with operations
 	specified using HA_ALTER_FLAGS and Alter_inplace_information.
@@ -332,7 +324,7 @@ public:
 	*/
 	bool inplace_alter_table(
 		TABLE*			altered_table,
-		Alter_inplace_info*	ha_alter_info);
+		Alter_inplace_info*	ha_alter_info) override;
 
 	/** Commit or rollback the changes made during
 	prepare_inplace_alter_table() and inplace_alter_table() inside
@@ -351,12 +343,12 @@ public:
 	bool commit_inplace_alter_table(
 		TABLE*			altered_table,
 		Alter_inplace_info*	ha_alter_info,
-		bool			commit);
+		bool			commit) override;
 	/** @} */
 
 	bool check_if_incompatible_data(
 		HA_CREATE_INFO*		info,
-		uint			table_changes);
+		uint			table_changes) override;
 
 	/** @name Multi Range Read interface @{ */
 
@@ -371,11 +363,11 @@ public:
 		void*			seq_init_param,
 		uint			n_ranges,
 		uint			mode,
-		HANDLER_BUFFER*		buf);
+		HANDLER_BUFFER*		buf) override;
 
 	/** Process next multi range read @see DsMrr_impl::dsmrr_next
 	@param range_info */
-	int multi_range_read_next(range_id_t *range_info);
+	int multi_range_read_next(range_id_t *range_info) override;
 
 	/** Initialize multi range read and get information.
 	@see ha_myisam::multi_range_read_info_const
@@ -394,7 +386,7 @@ public:
 		uint			n_ranges,
 		uint*			bufsz,
 		uint*			flags,
-		Cost_estimate*		cost);
+		Cost_estimate*		cost) override;
 
 	/** Initialize multi range read and get information.
 	@see DsMrr_impl::dsmrr_info
@@ -407,16 +399,16 @@ public:
 	@param cost */
 	ha_rows multi_range_read_info(uint keyno, uint n_ranges, uint keys,
 				      uint key_parts, uint* bufsz, uint* flags,
-				      Cost_estimate* cost);
+				      Cost_estimate* cost) override;
 
 	int multi_range_read_explain_info(uint mrr_mode,
-					  char *str, size_t size);
+					  char *str, size_t size) override;
 
 	/** Attempt to push down an index condition.
 	@param[in] keyno MySQL key number
 	@param[in] idx_cond Index condition to be checked
 	@return idx_cond if pushed; NULL if not pushed */
-	Item* idx_cond_push(uint keyno, Item* idx_cond);
+	Item* idx_cond_push(uint keyno, Item* idx_cond) override;
 	/* @} */
 
 	/** Check if InnoDB is not storing virtual column metadata for a table.
@@ -427,13 +419,35 @@ public:
 		return s.frm_version<FRM_VER_EXPRESSSIONS && s.virtual_fields;
 	}
 
-protected:
-	/**
-	MySQL calls this method at the end of each statement. This method
-	exists for readability only, called from reset(). The name reset()
-	doesn't give any clue that it is called at the end of a statement. */
-	int end_stmt();
+	/** Push a primary key filter.
+	@param[in]	pk_filter	filter against which primary keys
+					are to be checked
+	@retval	false if pushed (always) */
+	bool rowid_filter_push(Rowid_filter *rowid_filter) override;
 
+	bool
+	can_convert_string(const Field_string* field,
+			   const Column_definition& new_field) const override;
+	bool can_convert_varstring(
+	    const Field_varstring* field,
+	    const Column_definition& new_field) const override;
+	bool
+	can_convert_blob(const Field_blob* field,
+			 const Column_definition& new_field) const override;
+
+	/** @return whether innodb_strict_mode is active */
+	static bool is_innodb_strict_mode(THD* thd);
+
+	/** @return whether innodb_strict_mode is active */
+	bool is_innodb_strict_mode()
+	{ return is_innodb_strict_mode(m_user_thd); }
+	Compare_keys
+	compare_key_parts(const Field& old_field,
+			  const Column_definition& new_field,
+			  const KEY_PART_INFO& old_part,
+			  const KEY_PART_INFO& new_part) const override;
+
+protected:
 	dberr_t innobase_get_autoinc(ulonglong* value);
 	dberr_t innobase_lock_autoinc();
 	ulonglong innobase_peek_autoinc();
@@ -443,20 +457,22 @@ protected:
 	@see build_template() */
 	void reset_template();
 
+	/** @return whether the table is read-only */
+	bool is_read_only(bool altering_to_supported= false) const;
+
 	inline void update_thd(THD* thd);
 	void update_thd();
 
 	int general_fetch(uchar* buf, uint direction, uint match_mode);
 	int change_active_index(uint keynr);
-	/* @return true if it's necessary to switch current statement log
-	format from STATEMENT to ROW if binary log format is MIXED and
-	autoincrement values are changed in the statement */
-	bool autoinc_lock_mode_stmt_unsafe() const;
 	dict_index_t* innobase_get_index(uint keynr);
 
 #ifdef WITH_WSREP
-	int wsrep_append_keys(THD *thd, wsrep_key_type key_type,
-			      const uchar* record0, const uchar* record1);
+	int wsrep_append_keys(
+		THD *thd,
+		Wsrep_service_key_type key_type,
+		const uchar* record0,
+		const uchar* record1);
 #endif
 	/** Builds a 'template' to the prebuilt struct.
 
@@ -466,7 +482,7 @@ protected:
 	false if accessing individual fields is enough */
 	void build_template(bool whole_row);
 
-	virtual int info_low(uint, bool);
+	int info_low(uint, bool);
 
 	/** The multi range read session object */
 	DsMrr_impl		m_ds_mrr;
@@ -513,26 +529,12 @@ the definitions are bracketed with #ifdef INNODB_COMPATIBILITY_HOOKS */
 
 extern "C" {
 
-/** Check if a user thread is a replication slave thread
-@param thd user thread
-@retval 0 the user thread is not a replication slave thread
-@retval 1 the user thread is a replication slave thread */
-int thd_slave_thread(const MYSQL_THD thd);
-
 /** Check if a user thread is running a non-transactional update
 @param thd user thread
 @retval 0 the user thread is not running a non-transactional update
 @retval 1 the user thread is running a non-transactional update */
 int thd_non_transactional_update(const MYSQL_THD thd);
 
-/** Get high resolution timestamp for the current query start time.
-The timestamp is not anchored to any specific point in time,
-but can be used for comparison.
-@param thd user thread
-@retval timestamp in microseconds precision
-*/
-unsigned long long thd_start_utime(const MYSQL_THD thd);
-
 /** Get the user thread's binary logging format
 @param thd user thread
 @return Value to be used as index into the binlog_format_names array */
@@ -563,23 +565,8 @@ extern void mysql_bin_log_commit_pos(THD *thd, ulonglong *out_pos, const char **
 
 struct trx_t;
 #ifdef WITH_WSREP
-//extern "C" int wsrep_trx_order_before(void *thd1, void *thd2);
-
-extern "C" bool wsrep_thd_is_wsrep_on(THD *thd);
-
-
-extern "C" void wsrep_thd_set_exec_mode(THD *thd, enum wsrep_exec_mode mode);
-extern "C" void wsrep_thd_set_query_state(
-	THD *thd, enum wsrep_query_state state);
-
-extern "C" void wsrep_thd_set_trx_to_replay(THD *thd, uint64 trx_id);
-
-extern "C" uint32 wsrep_thd_wsrep_rand(THD *thd);
-extern "C" time_t wsrep_thd_query_start(THD *thd);
-extern "C" query_id_t wsrep_thd_query_id(THD *thd);
-extern "C" query_id_t wsrep_thd_wsrep_last_query_id(THD *thd);
-extern "C" void wsrep_thd_set_wsrep_last_query_id(THD *thd, query_id_t id);
-#endif
+#include <mysql/service_wsrep.h>
+#endif /* WITH_WSREP */
 
 extern const struct _ft_vft ft_vft_result;
 
@@ -647,6 +634,9 @@ public:
 	/** Set m_tablespace_type. */
 	void set_tablespace_type(bool table_being_altered_is_file_per_table);
 
+	/** Create InnoDB foreign keys from MySQL alter_info. */
+	dberr_t create_foreign_keys();
+
 	/** Create the internal innodb table.
 	@param create_fk	whether to add FOREIGN KEY constraints */
 	int create_table(bool create_fk = true);
@@ -716,26 +706,9 @@ public:
 	const char* table_name() const
 	{ return(m_table_name); }
 
-	/** @return whether the table needs to be dropped on rollback */
-	bool drop_before_rollback() const { return m_drop_before_rollback; }
-
 	THD* thd() const
 	{ return(m_thd); }
 
-	/** Normalizes a table name string.
-	A normalized name consists of the database name catenated to '/' and
-	table name. An example: test/mytable. On Windows normalization puts
-	both the database name and the table name always to lower case if
-	"set_lower_case" is set to true.
-	@param[in,out]	norm_name	Buffer to return the normalized name in.
-	@param[in]	name		Table name string.
-	@param[in]	set_lower_case	True if we want to set name to lower
-					case. */
-	static void normalize_table_name_low(
-		char*           norm_name,
-		const char*     name,
-		ibool           set_lower_case);
-
 private:
 	/** Parses the table name into normal name and either temp path or
 	remote path if needed.*/
@@ -765,8 +738,6 @@ private:
 	char*		m_table_name;
 	/** Table */
 	dict_table_t*	m_table;
-	/** Whether the table needs to be dropped before rollback */
-	bool		m_drop_before_rollback;
 
 	/** Remote path (DATA DIRECTORY) or zero length-string */
 	char*		m_remote_path;
@@ -869,15 +840,6 @@ innodb_base_col_setup_for_stored(
 /** whether this is a stored generated column */
 #define innobase_is_s_fld(field) ((field)->vcol_info && (field)->stored_in_db())
 
-/** Always normalize table name to lower case on Windows */
-#ifdef _WIN32
-#define normalize_table_name(norm_name, name)           \
-	create_table_info_t::normalize_table_name_low(norm_name, name, TRUE)
-#else
-#define normalize_table_name(norm_name, name)           \
-	create_table_info_t::normalize_table_name_low(norm_name, name, FALSE)
-#endif /* _WIN32 */
-
 /** Converts a search mode flag understood by MySQL to a flag understood
 by InnoDB.
 @param[in]	find_flag	MySQL search mode flag.
@@ -946,10 +908,8 @@ innodb_col_no(const Field* field)
 /********************************************************************//**
 Helper function to push frm mismatch error to error log and
 if needed to sql-layer. */
-UNIV_INTERN
 void
 ib_push_frm_error(
-/*==============*/
 	THD*		thd,		/*!< in: MySQL thd */
 	dict_table_t*	ib_table,	/*!< in: InnoDB table */
 	TABLE*		table,		/*!< in: MySQL table */
