@@ -2,7 +2,7 @@
 
 Copyright (c) 2010, 2015, Oracle and/or its affiliates. All Rights Reserved.
 Copyright (c) 2012, Facebook Inc.
-Copyright (c) 2013, 2020, MariaDB Corporation.
+Copyright (c) 2013, 2021, MariaDB Corporation.
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the GNU General Public License as published by the
@@ -36,7 +36,7 @@ Created 12/15/2009	Jimmy Yang
 #define __STDC_LIMIT_MACROS
 #endif /* __STDC_LIMIT_MACROS */
 
-#include <stdint.h>
+#include <cstdint>
 #include "my_atomic.h"
 #include "my_atomic_wrapper.h"
 
@@ -136,8 +136,6 @@ enum monitor_id_t {
 	/* Start of Metadata counter */
 	MONITOR_MODULE_METADATA,
 	MONITOR_TABLE_OPEN,
-	MONITOR_TABLE_CLOSE,
-	MONITOR_TABLE_REFERENCE,
 
 	/* Lock manager related counters */
 	MONITOR_MODULE_LOCK,
@@ -195,18 +193,9 @@ enum monitor_id_t {
 	MONITOR_FLUSH_N_TO_FLUSH_REQUESTED,
 
 	MONITOR_FLUSH_N_TO_FLUSH_BY_AGE,
-	MONITOR_FLUSH_ADAPTIVE_AVG_TIME_SLOT,
-	MONITOR_LRU_BATCH_FLUSH_AVG_TIME_SLOT,
-
-	MONITOR_FLUSH_ADAPTIVE_AVG_TIME_THREAD,
-	MONITOR_LRU_BATCH_FLUSH_AVG_TIME_THREAD,
-	MONITOR_FLUSH_ADAPTIVE_AVG_TIME_EST,
-	MONITOR_LRU_BATCH_FLUSH_AVG_TIME_EST,
-	MONITOR_FLUSH_AVG_TIME,
+	MONITOR_FLUSH_ADAPTIVE_AVG_TIME,
 
 	MONITOR_FLUSH_ADAPTIVE_AVG_PASS,
-	MONITOR_LRU_BATCH_FLUSH_AVG_PASS,
-	MONITOR_FLUSH_AVG_PASS,
 
 	MONITOR_LRU_GET_FREE_LOOPS,
 	MONITOR_LRU_GET_FREE_WAITS,
@@ -229,14 +218,7 @@ enum monitor_id_t {
 	MONITOR_LRU_BATCH_SCANNED_NUM_CALL,
 	MONITOR_LRU_BATCH_SCANNED_PER_CALL,
 	MONITOR_LRU_BATCH_FLUSH_TOTAL_PAGE,
-	MONITOR_LRU_BATCH_FLUSH_COUNT,
-	MONITOR_LRU_BATCH_FLUSH_PAGES,
 	MONITOR_LRU_BATCH_EVICT_TOTAL_PAGE,
-	MONITOR_LRU_BATCH_EVICT_COUNT,
-	MONITOR_LRU_BATCH_EVICT_PAGES,
-	MONITOR_LRU_SINGLE_FLUSH_SCANNED,
-	MONITOR_LRU_SINGLE_FLUSH_SCANNED_NUM_CALL,
-	MONITOR_LRU_SINGLE_FLUSH_SCANNED_PER_CALL,
 	MONITOR_LRU_SINGLE_FLUSH_FAILURE_COUNT,
 	MONITOR_LRU_GET_FREE_SEARCH,
 	MONITOR_LRU_SEARCH_SCANNED,
@@ -301,7 +283,6 @@ enum monitor_id_t {
 	MONITOR_TRX_COMMIT_UNDO,
 	MONITOR_TRX_ROLLBACK,
 	MONITOR_TRX_ROLLBACK_SAVEPOINT,
-	MONITOR_TRX_ACTIVE,
 	MONITOR_RSEG_HISTORY_LEN,
 	MONITOR_NUM_UNDO_SLOT_USED,
 	MONITOR_NUM_UNDO_SLOT_CACHED,
@@ -326,7 +307,6 @@ enum monitor_id_t {
 	MONITOR_LSN_CHECKPOINT_AGE,
 	MONITOR_OVLD_BUF_OLDEST_LSN,
 	MONITOR_OVLD_MAX_AGE_ASYNC,
-	MONITOR_OVLD_MAX_AGE_SYNC,
 	MONITOR_PENDING_LOG_FLUSH,
 	MONITOR_PENDING_CHECKPOINT_WRITE,
 	MONITOR_LOG_IO,
@@ -365,9 +345,7 @@ enum monitor_id_t {
 	/* Adaptive Hash Index related counters */
 	MONITOR_MODULE_ADAPTIVE_HASH,
 	MONITOR_OVLD_ADAPTIVE_HASH_SEARCH,
-#endif /* BTR_CUR_HASH_ADAPT */
 	MONITOR_OVLD_ADAPTIVE_HASH_SEARCH_BTREE,
-#ifdef BTR_CUR_HASH_ADAPT
 	MONITOR_ADAPTIVE_HASH_PAGE_ADDED,
 	MONITOR_ADAPTIVE_HASH_PAGE_REMOVED,
 	MONITOR_ADAPTIVE_HASH_ROW_ADDED,
@@ -397,27 +375,13 @@ enum monitor_id_t {
 	MONITOR_OVLD_SERVER_ACTIVITY,
 	MONITOR_MASTER_ACTIVE_LOOPS,
 	MONITOR_MASTER_IDLE_LOOPS,
-	MONITOR_SRV_BACKGROUND_DROP_TABLE_MICROSECOND,
-	MONITOR_SRV_IBUF_MERGE_MICROSECOND,
 	MONITOR_SRV_LOG_FLUSH_MICROSECOND,
-	MONITOR_SRV_MEM_VALIDATE_MICROSECOND,
-	MONITOR_SRV_PURGE_MICROSECOND,
 	MONITOR_SRV_DICT_LRU_MICROSECOND,
 	MONITOR_SRV_DICT_LRU_EVICT_COUNT_ACTIVE,
 	MONITOR_SRV_DICT_LRU_EVICT_COUNT_IDLE,
-	MONITOR_SRV_CHECKPOINT_MICROSECOND,
 	MONITOR_OVLD_SRV_DBLWR_WRITES,
 	MONITOR_OVLD_SRV_DBLWR_PAGES_WRITTEN,
 	MONITOR_OVLD_SRV_PAGE_SIZE,
-	MONITOR_OVLD_RWLOCK_S_SPIN_WAITS,
-	MONITOR_OVLD_RWLOCK_X_SPIN_WAITS,
-	MONITOR_OVLD_RWLOCK_SX_SPIN_WAITS,
-	MONITOR_OVLD_RWLOCK_S_SPIN_ROUNDS,
-	MONITOR_OVLD_RWLOCK_X_SPIN_ROUNDS,
-	MONITOR_OVLD_RWLOCK_SX_SPIN_ROUNDS,
-	MONITOR_OVLD_RWLOCK_S_OS_WAITS,
-	MONITOR_OVLD_RWLOCK_X_OS_WAITS,
-	MONITOR_OVLD_RWLOCK_SX_OS_WAITS,
 
 	/* Data DML related counters */
 	MONITOR_MODULE_DML_STATS,
@@ -433,7 +397,6 @@ enum monitor_id_t {
 	/* Data DDL related counters */
 	MONITOR_MODULE_DDL_STATS,
 	MONITOR_BACKGROUND_DROP_INDEX,
-	MONITOR_BACKGROUND_DROP_TABLE,
 	MONITOR_ONLINE_CREATE_INDEX,
 	MONITOR_PENDING_ALTER_TABLE,
 	MONITOR_ALTER_TABLE_SORT_FILES,
@@ -445,10 +408,6 @@ enum monitor_id_t {
 	MONITOR_ICP_OUT_OF_RANGE,
 	MONITOR_ICP_MATCH,
 
-	/* Mutex/RW-Lock related counters */
-	MONITOR_MODULE_LATCHES,
-	MONITOR_LATCHES,
-
 	/* This is used only for control system to turn
 	on/off and reset all monitor counters */
 	MONITOR_ALL_COUNTER,
@@ -654,14 +613,14 @@ Use MONITOR_DEC if appropriate mutex protection exists.
 		}							\
 	}
 
-#ifdef HAVE_valgrind
+#ifdef HAVE_MEM_CHECK
 # define MONITOR_CHECK_DEFINED(value) do {	\
     mon_type_t m __attribute__((unused))= value;        \
 	MEM_CHECK_DEFINED(&m, sizeof m);	\
 } while (0)
-#else /* HAVE_valgrind */
+#else /* HAVE_MEM_CHECK */
 # define MONITOR_CHECK_DEFINED(value) (void) 0
-#endif /* HAVE_valgrind */
+#endif /* HAVE_MEM_CHECK */
 
 #define	MONITOR_INC_VALUE(monitor, value)				\
 	MONITOR_CHECK_DEFINED(value);					\
